
--[[

Puzzlet

collisions.lua

(c) bob fossil 2016

--]]

TRAP_COLLISION_MOVEABLE = 255
TRAP_COLLISION_WEAPON = 256

TRAP_INVALID = 0

COLLISION_NONE = 0
COLLISION_WALL = 1
COLLISION_DOOR = 2
COLLISION_PLAYER = 3
COLLISION_UNCOLLECTABLE_OBJECT = 4
COLLISION_EXIT = 5
COLLISION_TRAP = 6
COLLISION_ENEMY = 7
COLLISION_GENERATOR = 8
COLLISION_KEY = 9
COLLISION_POTION = 10
COLLISION_FORCEFIELD = 11
COLLISION_MOVEABLE_WALL = 12
COLLISION_FOOD = 13

COLLSION_WRAPPED = 100

-- Returns true if rect1 intersects with rect2 or false if it doesn't.
function rect_intersect_rect(rect1, rect2)

	if rect1.left < rect1.right and rect1.top < rect1.bottom and 		-- rect1 valid
		rect2.left < rect2.right and rect2.top < rect2.bottom and	-- rect2 valid
		rect1.left < rect2.right and rect2.left < rect1.right and	-- intersects horizontally
		rect1.top < rect2.bottom and rect2.top < rect1.bottom then	-- intersects vertically

		return true
	end

	return false

end

function rect_equal_rect(rect1, rect2)

	if rect1.left==rect2.left and rect1.top==rect2.top and
		rect1.right==rect2.right and rect1.bottom==rect2.bottom then
		return true
	end

	return false

end

-- Collison handlers

-- Returns what we collided with or COLLISION_NONE.
function check_for_tile_collision(map_x, map_y)

	local collision = COLLISION_NONE

	-- Returns true if we can move the player into the specified tile.
	local value = global.map[map_y][map_x]
	
	--dprintf("%d %d %d", map_y, map_x, value)
	
	-- Bail out if the tile is in the walls or door range.
	if value > LEVEL_EXIT_RANGE then
		if value > LEVEL_SPECIAL_RANGE then
			if value < LEVEL_PLAYER_RANGE then

				value = value - LEVEL_SPECIAL_RANGE
			
				-- Special range.
				-- COLLISION_KEY and COLLISION_POTION are handled
				-- by the dynamic collision_handler.
				if value==GRAPHIC_SPECIAL_BONES then
					collision = COLLISION_PLAYER
				
				elseif value==GRAPHIC_SPECIAL_GENERATOR_DAMAGED or
					value==GRAPHIC_SPECIAL_GENERATOR_OFF or
					value==GRAPHIC_SPECIAL_GENERATOR_ON then
					
					collision = COLLISION_GENERATOR
				
				elseif value==GRAPHIC_SPECIAL_GROUND_FORCEFIELD then
					collision = COLLISION_FORCEFIELD
					
				end
			end
			
		-- COLLISION_EXIT is handled by the dynamic collision_handler.
		end

	elseif value > LEVEL_DOOR_RANGE then
		collision = COLLISION_DOOR
		
	else
		if value==GRAPHIC_WALL_MOVEABLE then
			collision = COLLISION_MOVEABLE_WALL
		else
			collision = COLLISION_WALL
		end
	end

	--dprintf("collision %d", collision)
	
	return collision
end

-- General purpose collision detection routine for moving things.
-- Returns what we collided with and it's index or COLLISION_NONE.
function check_for_dynamic_collisions(rect, player_index, enemy_index)

	local collision = COLLISION_NONE

	-- Check for collision with another player.
	for player = 1, #global.players do
		-- Don't check ourself.
		if player~=player_index then
			if global.players[player].active and
				not global.players[player].exited then
					
				if rect_intersect_rect(global.players[player], rect) then	
					collision = COLLISION_PLAYER
					collision_index = player
					break
				end
			end
		end
	end
	
	-- Check for collision with enemies.
	if collision==COLLISION_NONE then
		for enemy = 1, #global.enemies do
		-- Don't check ourself.	
			if enemy~=enemy_index and
				rect_intersect_rect(global.enemies[enemy], rect) then

				collision = COLLISION_ENEMY
				collision_index = enemy
				break;
			end
		end
	end
	
	-- Check for collision with objects.
	if collision==COLLISION_NONE then
		for object = 1, #global.level_objects do
	
			if not global.level_objects[object].collected and
				rect_intersect_rect(rect, global.level_objects[object]) then
			
				--dprintf("mv: %d map_y %d map_x %d - %d %d",  map_value, map_y, map_x, player_rect.left, player_rect.top)
			
				if global.level_objects[object]._type==OBJECT_TYPE_KEY then
			
					collision = COLLISION_KEY
					collision_index = object
					break;
				
				elseif global.level_objects[object]._type==OBJECT_TYPE_POTION then

					collision = COLLISION_POTION
					collision_index = object
					break;
					
				elseif global.level_objects[object]._type==OBJECT_TYPE_FOOD then

					collision = COLLISION_FOOD
					collision_index = object
					break;					
				end
			end
		end
	end
	
	-- Check for collision with exits.
	if collision==COLLISION_NONE then
		for _exit = 1, #global.exits do
		
			local exit_table = global.exits[_exit]
	
			if rect_intersect_rect(rect, exit_table) then
			
				local left_diff = math.abs(rect.left - exit_table.left)
				local top_diff = math.abs(rect.top - exit_table.top)

				--dprintf("left_diff %d top_diff %d", left_diff, top_diff)
				
				if left_diff <= 10 and top_diff <= 10 then
					collision = COLLISION_EXIT
					collision_index = _exit
				end
			end
		end
	end
	
	return collision, collision_index

end

--

-- Returns true if we could move the block in the desired location.
function handle_moveable_wall(map_x, map_y, direction)

	local moved = false

	--dprintf("Warrior hit moveable wall!")

	local new_map_x = map_x
	local new_map_y = map_y
	
	-- Work out which direction to push.
	if direction==UP then
		new_map_y = new_map_y - 1
	elseif direction==DOWN then
		new_map_y = new_map_y + 1
	elseif direction==LEFT then
		new_map_x = new_map_x - 1
	elseif direction==RIGHT then
		new_map_x = new_map_x + 1
	else
		-- !!! You can't move a wall on a diagonal.
		return moved
	end
	
	local new_map_rect = {left = (new_map_x - 1) * TILE_WIDTH,
		top = (new_map_y - 1) * TILE_HEIGHT}
	new_map_rect.right = new_map_rect.left + (TILE_WIDTH - 1)
	new_map_rect.bottom = new_map_rect.top + (TILE_HEIGHT - 1)
	
	local wrapped, wrapped_rect = handle_wrap_around(new_map_rect, direction)
	if wrapped then
		new_map_x = (wrapped_rect.left / TILE_WIDTH) + 1
		new_map_y = (wrapped_rect.top / TILE_HEIGHT) + 1
		
		new_map_rect.left = wrapped_rect.left
		new_map_rect.top = wrapped_rect.top
		new_map_rect.right = wrapped_rect.right
		new_map_rect.bottom = wrapped_rect.bottom		
	end
	
	-- Check against current player and enemy positions.
	for player = 1, #global.players do
	
		if rect_intersect_rect(global.players[player], new_map_rect) then

			if global.players[player].active and
				global.players[player].alive and
				global.players[player].exited==false then

				return moved
			end
		end
	end
	
	for enemy = 1, #global.enemies do
	
		if rect_intersect_rect(global.enemies[enemy], new_map_rect) then
			return moved
		end
	end
	
	local check = global.map[new_map_y][new_map_x]

	-- We can only move a wall piece if there is free space in the direction we're pushing.	
	if check == GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE then
		global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
		global.map[new_map_y][new_map_x] = GRAPHIC_WALL_MOVEABLE + LEVEL_WALL_RANGE
		
		-- We've moved a block. Check if this affects any trap tiles which may have blocks on them.
		
		local block_left = (map_x - 1) * TILE_WIDTH
		local block_top = (map_y - 1) * TILE_HEIGHT
	
		for trap = 1, #global.traps do

			if global.traps[trap].player==TRAP_COLLISION_MOVEABLE then
			
				if block_left==global.traps[trap].left and
					block_top==global.traps[trap].top then

					handle_trap_collisions(trap, false, 0, false)
					global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND_TRAP + LEVEL_SPECIAL_RANGE
				end
			end
		end

		moved = true

	elseif check == GRAPHIC_SPECIAL_GROUND_TRAP + LEVEL_SPECIAL_RANGE then

		-- Free space we're moving onto is a trap tile.
		--dprintf("moveable block onto trap tile")
		
		local block_left = (new_map_x - 1) * TILE_WIDTH
		local block_top = (new_map_y - 1) * TILE_HEIGHT
--		local block_rect = {left = (check_x - 1) * TILE_WIDTH, top = (check_y - 1) * TILE_HEIGHT,
--			right = check_x * TILE_WIDTH, bottom = check_y * TILE_HEIGHT}

		-- Find our trap tile.
		for trap = 1, #global.traps do

--			if rect_equal_rect(block_rect, global.traps[trap]) then
			if block_left==global.traps[trap].left and block_top==global.traps[trap].top then

				-- Spring the trap.
				handle_trap_collisions(trap, true, 0, false)
				
				-- Move the block to the trap position.
				global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
				global.map[new_map_y][new_map_x] = GRAPHIC_WALL_MOVEABLE + LEVEL_WALL_RANGE
				-- Set special case that the trap has been activated by a moveable wall.
				global.traps[trap].player = TRAP_COLLISION_MOVEABLE
			end
		end
		
	elseif (check>=GRAPHIC_EXIT_1 + LEVEL_EXIT_RANGE or
		check <= GRAPHIC_EXIT_8 + LEVEL_EXIT_RANGE) then

		-- Free space we're moving onto is a teleport exit.		
		local exit_type = (check - LEVEL_EXIT_RANGE) - 1
		
		for area = 1, #global.areas do
			if exit_type==global.areas[area].id then
			
				local teleport_ok = can_teleport(area)
			
				if teleport_ok then
				
					global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
					local tx = (global.areas[area].left / TILE_WIDTH) + 1
					local ty = (global.areas[area].top / TILE_HEIGHT) + 1
					global.map[ty][tx] = GRAPHIC_WALL_MOVEABLE + LEVEL_WALL_RANGE
				
				end	
			end
		end
		
	end
		
	return moved
end

function handle_tile_from_trap(tx, ty, tile_value, trap, apply)

	--dprintf("Set %d %d to %d", ty, tx, tile_value)
		
	global.map[ty][tx] = tile_value

	-- If we add certain tiles from a trap, update the level structures as well.

	-- !!! Exits? Exit destinations?
	-- global.exits global.areas
	
	if tile_value==GRAPHIC_SPECIAL_KEY + LEVEL_SPECIAL_RANGE then
		-- Key
		local objects = #global.level_objects
		global.level_objects[objects + 1] = {left = (tx - 1) * TILE_WIDTH, top = (ty - 1) * TILE_HEIGHT,
			right = (tx * TILE_WIDTH), bottom = (ty * TILE_HEIGHT),
			collected = false, _type = OBJECT_TYPE_KEY, from_trap = trap}
		global.traps[trap].tiles[apply].object = objects + 1

	elseif tile_value==GRAPHIC_SPECIAL_POTION_KILL_ALL + LEVEL_SPECIAL_RANGE or
		tile_value==GRAPHIC_SPECIAL_POTION_KILL_DEATH + LEVEL_SPECIAL_RANGE or
		tile_value==GRAPHIC_SPECIAL_POTION_KILL_ENEMIES + LEVEL_SPECIAL_RANGE then	

		global.map[ty][tx] = GRAPHIC_SPECIAL_POTION + LEVEL_SPECIAL_RANGE
		
		-- Potion
		local potion = (tile_value - (GRAPHIC_SPECIAL_POTION_KILL_ALL + LEVEL_SPECIAL_RANGE)) + 1
		local objects = #global.level_objects
		global.level_objects[objects + 1] = {left = (tx - 1) * TILE_WIDTH, top = (ty - 1) * TILE_HEIGHT,
			right = (tx * TILE_WIDTH), bottom = (ty * TILE_HEIGHT),
			collected = false, _type = OBJECT_TYPE_POTION, potion_type = potion, from_trap = trap}
		global.traps[trap].tiles[apply].object = objects + 1

	elseif tile_value==GRAPHIC_SPECIAL_FOOD_ALE + LEVEL_SPECIAL_RANGE or
		tile_value==GRAPHIC_SPECIAL_FOOD_MEAT + LEVEL_SPECIAL_RANGE then
	
		-- Food
		local food = (tile_value - (GRAPHIC_SPECIAL_FOOD_ALE + LEVEL_SPECIAL_RANGE)) + 1		
		local objects = #global.level_objects
		global.level_objects[objects + 1] = {left = (tx - 1) * TILE_WIDTH, top = (ty - 1) * TILE_HEIGHT,
			right = (tx * TILE_WIDTH), bottom = (ty * TILE_HEIGHT),
			collected = false, _type = OBJECT_TYPE_FOOD, food_type = food, from_trap = trap}
		global.traps[trap].tiles[apply].object = objects + 1
		
	elseif tile_value > LEVEL_EXIT_RANGE and tile_value < LEVEL_SPECIAL_RANGE then

		-- Exit
		local exit = (tile_value - LEVEL_EXIT_RANGE) - 1
		local exits = #global.exits
		global.exits[exits + 1] = {left = (tx - 1) * TILE_WIDTH, top = (ty - 1) * TILE_HEIGHT,
			right = (tx * TILE_WIDTH), bottom = (ty * TILE_HEIGHT), _type = exit, state = EXIT_STATE_STABLE, from_trap = trap}
		global.traps[trap].tiles[apply].exit = exits + 1

	elseif tile_value>=LEVEL_ENEMY_RANGE then
		-- Enemy
		local enemies = #global.enemies
		add_enemy(global.traps[trap].tiles[apply].tile_index, (tx  - 1) * TILE_WIDTH, (ty - 1) * TILE_HEIGHT, 0)
		-- Put enemy on safe ground.
		global.map[ty][tx] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
		global.traps[trap].tiles[apply].enemy = enemies + 1
	end

end

-- !!! Mark player as dead if trap puts a wall on top of them when applied or restored.

function apply_trap(trap, restore, copy)

	for apply = 1, #global.traps[trap].tiles do

		local tx = global.traps[trap].tiles[apply].map_x
		local ty = global.traps[trap].tiles[apply].map_y

		if restore then
			-- Restore original tile.
			global.map[ty][tx] = global.traps[trap].tiles[apply].copy 
			global.traps[trap].tiles[apply].copy = nil
			-- We have to remove certain added tiles from tables too.
			if global.traps[trap].tiles[apply].object then
				table.remove(global.level_objects, global.traps[trap].tiles[apply].object)
				global.traps[trap].tiles[apply].object = nil
			elseif global.traps[trap].tiles[apply].exit then
				table.remove(global.exits, global.traps[trap].tiles[apply].exit)
				global.traps[trap].tiles[apply].exit = nil
			end
		else
			if copy then
				-- Store original value.
				global.traps[trap].tiles[apply].copy = global.map[ty][tx]	
			end

			-- Apply the new tile.
			local tile_value = (global.traps[trap].tiles[apply].tile_category) +
				global.traps[trap].tiles[apply].tile_index
			
			handle_tile_from_trap(tx, ty, tile_value, trap, apply)
				
		end
	end

end

function handle_trap_collisions(trap, in_trap, collision, from_player)

	local trap_collision = collision

	if global.traps[trap].activated==false then
	
		if in_trap==true then
		
			local update = false
			global.traps[trap].activated = true

			if global.traps[trap].trap_type==TRAP_ONCE then

				--dprintf("Trap %d activated", trap)
			
				apply_trap(trap, false, false)
				
				-- Remove this trap.
				local map_x = (global.traps[trap].left / TILE_WIDTH) + 1
				local map_y = (global.traps[trap].top / TILE_HEIGHT) + 1

				-- Replace trap tile with ground tile.
				global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
				global.traps[trap].trap_type = TRAP_INVALID
				--update = true
				
			elseif global.traps[trap].trap_type==TRAP_SWITCH then
			
				if global.traps[trap].on==true then

					--dprintf("Trap %d deactivated", trap)
				
					global.traps[trap].on = false
					apply_trap(trap, true, false)
					global.traps[trap].player = 0
					
				else
					--dprintf("Trap %d activated", trap)
				
					global.traps[trap].on = true
					apply_trap(trap, false, true)
					
					update = true

				end
			
			elseif global.traps[trap].trap_type==TRAP_PRESSURE then

				--dprintf("Trap %d activated", trap)
					
				apply_trap(trap, false, true)
				
				update = true
			
			end

			if update==true and from_player==true then
				-- Position player onto the trap tile.
				global.players[global.current_player].left = global.traps[trap].left
				global.players[global.current_player].top = global.traps[trap].top
				global.players[global.current_player].right = global.traps[trap].left + TILE_WIDTH
				global.players[global.current_player].bottom = global.traps[trap].top + TILE_HEIGHT
				-- Remember which player activated the trap.
				global.traps[trap].player = global.current_player
			
				trap_collision = COLLISION_TRAP
			end
		end
	else
		-- Tile is already activated.
		-- Are we still on it?
		if in_trap==false then
			-- Check if it's a pressure trap. Deactivate if we're no longer standing
			-- on it.
			if global.traps[trap].trap_type==TRAP_PRESSURE then
		
				--dprintf("Trap %d deactivated", trap)
				
				apply_trap(trap, true, false)
				
			elseif global.traps[trap].trap_type==TRAP_SWITCH then
				-- Handle special case if switch is on and hit by a weapon.
				if global.traps[trap].player==TRAP_COLLISION_WEAPON and
					global.traps[trap].on==true then

					--dprintf("Trap %d deactivated", trap)
				
					global.traps[trap].on = false
					apply_trap(trap, true, false)
					global.traps[trap].player = 0
					
				end		
			end

			global.traps[trap].activated = false
			global.traps[trap].player = 0
		end
	end

	return trap_collision
end

